import static java.lang.System.out;
import static java.lang.Math.max;
import static java.lang.Math.min;

class TwoThreeTree{
	private INode23 _root;
	private EmptyTree23 _emptyTree;
	public INode23 getEmptyTree23(){
		if(_emptyTree == null){
			_emptyTree = new EmptyTree23();
		} 
		return _emptyTree;
	}
	public TwoThreeTree(){
		_root = getEmptyTree23();
	}
	private INode23 findNode(int key){
		INode23 root = _root;
		while (!root.testNode(key)){
			root = root.nextNode(key);
		}
		return root;
	}
	public void insert(int key){
		findNode(key).insertValue(key);
	}
	@Override 
	public String toString(){
		return _root.toString();
	}

	///////////// INode23
	
	private interface INode23 {	
		boolean isEmpty();
		boolean testNode(int key); //should we insert new node here
		INode23 nextNode(int key); // next node when serching where to insert a new node
		INode23 insertNode(Node2 node); // insert new node here
		INode23 insertValue(int key); // insert Value here
		INode23 getParent();
		void setParent(INode23 newParent);
		INode23 replaceChild(INode23 oldNode, INode23 newNode);		
	}
	
	//////////// Node23
	
	abstract class Node23 implements INode23{
		
		protected INode23 _parent;
		
		@Override
		public boolean isEmpty(){
			return false;
		}
		
		@Override	
		public INode23 getParent(){
			return _parent;
		}
	
		@Override	
		public void setParent(INode23 newParent){
			_parent = newParent;
		}
	}
	
	//////////// EmptyTree23
	
	private class EmptyTree23 extends Node23{
		
		private EmptyTree23(){} 
		
		@Override
		public boolean isEmpty(){
			return true;
		}
		
		@Override 
		public INode23 insertValue(int key){
			out.println("Empty tree got "+key);
			_root = new Node2(key, getEmptyTree23());
			return _root;
		}
		
		@Override
		public INode23 insertNode(Node2 node){
			_root = node;
			node.setParent(getEmptyTree23());
			return node;
		}
	
		@Override
		public boolean testNode(int key){//We start inserting with the empty tree
			return true;
		}
	
		@Override
		public INode23 nextNode(int key){//could we make it smarter?
			return this;
		}
		
		@Override//the only child of the empty tree is the non-empty root 
		public INode23 replaceChild(INode23 oldNode, INode23 newNode){
			return oldNode == _root?_root=newNode:_root;
		}
		
		@Override 
		public String toString(){
			return "+";
		} 
	}
	
	//////////// Leaf23
		
		private class Leaf23 extends Node23{
			
			private Leaf23(){} 
			
			@Override
			public boolean isEmpty(){
				return true;
			}
			
			@Override 
			public INode23 insertValue(int key){
				return insertNode(new Node2(key));
			}
			
			@Override
			public INode23 insertNode(Node2 node){//Is this ever called?
				out.println("Inserting into leaf: "+node.getValue());				
				node.setParent(_parent);
				_parent.replaceChild(this, node);
				return node;
			}
		
			@Override
			public boolean testNode(int key){
				return true;
			}
		
			@Override
			public INode23 nextNode(int key){//could we make it smarter
				return this;
			}
			
			@Override//could we make it smarter?
			public INode23 replaceChild(INode23 oldNode, INode23 newNode){
				return this;
			}
			
			@Override 
			public String toString(){
				return "-";
			} 
		}
	
	//////////// Node2
	
	class Node2 extends Node23{
		INode23 _left, _right;
		int _value;
		
		public Node2(int key){
			this(key,null);//caution null
		}
		
		public Node2(int key, INode23 parent){
			this(key, parent, new Leaf23(), new Leaf23());
		}
		
		public Node2(int key, INode23 parent, INode23 left, INode23 right){
			_value = key;
			_parent = parent;
			_right = right; 
			_left = left;
			out.println("Node2 created "+key);
		}
		
		public void setRight(INode23 right){
			_right = right;
		}
		
		public void setLeft(INode23 left){
			_left = left; 
		}
		
		public INode23 getLeft(){
			return _left; 
		}
		
		public INode23 getRight(){
			return _right;
		}
		
		public int getValue(){
			return _value;
		} 
		
		@Override
		public boolean testNode(int key){
			out.println("Node2.testNode "+key);
			return _left.isEmpty()&&_right.isEmpty()||_value==key;
		}
		
		@Override
		public INode23 nextNode(int key){
			if(key>_value){
				return _right;
			} else {
				return _left;
			}
		}
		@Override 
		public INode23 insertValue(int key){
			return insertNode(new Node2(key));
		}
		
		@Override 
		public INode23 insertNode(Node2 node){
			int key = node.getValue();
			if(key<_value){
				return insertMin(node, key);
			} else {
				return insertMax(node, key);
			}
		}		
		 
		private INode23 insertMin(Node2 node, int key){
			out.println("Node2.insertMin "+key+" parent = "+getParent());		
			Node3 node3 = new Node3(key, _value, getParent());
			node3.setLeft(node.getLeft());
			node3.setMiddle(node.getRight());
			node3.setRight(getRight());
			node.getLeft().setParent(node3);
			node.getRight().setParent(node3);
			getRight().setParent(node3);
			return _parent.replaceChild(this, node3);
		}	
	
		private INode23 insertMax(Node2 node, int key){
			out.println("Node2.insertMax "+key+" parent = "+getParent());		
			Node3 node3 = new Node3(_value, key, getParent());
			node3.setLeft(getLeft());
			node3.setMiddle(node.getLeft());
			node3.setRight(node.getRight());
			node.getLeft().setParent(node3);
			node.getRight().setParent(node3);
			getLeft().setParent(node3);
			return _parent.replaceChild(this, node3);
		}
	
	//Caution: call only for _parent.replaceChild(this, ...);
		@Override
		public INode23 replaceChild(INode23 oldNode, INode23 newNode){
			if(_left == oldNode){
				_left = newNode;
			} else {
				_right = newNode;
			}
			return newNode;
		}
	
		@Override 
		public String toString(){
			return "["+_left+","+_value+","+_right+"]";
		}	
	}
	
	//////////// Node3
	
	class Node3 extends Node23{
		INode23 _left, _middle, _right;
		int _min, _max;
	
		public Node3(int minKey, int maxKey){
			this(minKey, maxKey, null);//caution null
			
		}
	
		public Node3(int minKey, int maxKey, INode23 parent){
			_min = minKey;
			_max = maxKey;
			_parent = parent;
			_middle = new Leaf23();
			_left =  new Leaf23();
			_right =  new Leaf23();
			out.println("Node3 created "+_min+"-"+_max);
		}
		
		public void setLeft(INode23 left){
			_left = left;
		}
		
		public void setMiddle(INode23 middle){
			_middle = middle;
		}
	
		public void setRight(INode23 right){
			_right = right;
		}
	
		@Override	
		public boolean testNode(int key){
			return _left.isEmpty()&&_right.isEmpty()&&_middle.isEmpty()
					||equalsNodeValue(key);
		}
		private boolean equalsNodeValue(int key){
			return key==_min||key==_max;
		}
	
		@Override
		public INode23 nextNode(int key){
			if(key<_min){
				return _left;
			} else {
				if(key<_max){
					return _middle;
				} else {
					return _right;
				}
			}
		}
		
		
		@Override 
		public INode23 insertValue(int key){
			return insertNode(new Node2(key));
		}
		
	 	@Override 
		public INode23 insertNode(Node2 node){
			int key = node.getValue();
			if (key<_min){
				return insertLeft(node);
			} else {
				if (key<_max){
					return insertMiddle(node);
				} else {
					return insertRight(node);
				}
			}
		}
	
		private INode23 insertLeft(Node2 node){
			out.println("Node3.insertLeft "+node.getValue());
			Node2 newRoot = new Node2(_min, _parent);
			Node2 newLeft = node;
			newLeft.setParent(newRoot); 
			Node2 newRight = new Node2(_max, newRoot);
			
			newRight.setLeft(_middle);
			newRight.setRight(_right);
			
			newRoot.setLeft(newLeft);
			newRoot.setRight(newRight);
	
			return _parent.insertNode(newRoot);
		}
		
		private INode23 insertMiddle(Node2 node){
			out.println("Node3.insertMiddle "+node.getValue());
			Node2 newRoot = node; 
			newRoot.setParent(_parent);
			Node2 newLeft = new Node2(_min, newRoot); 
			Node2 newRight = new Node2(_max, newRoot);
			
			newLeft.setLeft(_left);
			newLeft.setRight(node.getLeft());
			newRight.setLeft(node.getRight());
			newRight.setRight(_right);
			
			newRoot.setLeft(newLeft);
			newRoot.setRight(newRight);
	
			return _parent.insertNode(newRoot);
		}
		
		private INode23 insertRight(Node2 node){
			out.println("Node3.insertRight "+node.getValue());
			Node2 newRoot = new Node2(_max, _parent);
			Node2 newLeft = new Node2(_min, newRoot); 
			Node2 newRight = node; 
			newRight.setParent(newRoot);
			
			newLeft.setLeft(_left);
			newLeft.setRight(_middle);
			
			newRoot.setLeft(newLeft);
			newRoot.setRight(newRight);
	
			return _parent.insertNode(newRoot);
		}	
		
		
		//Caution: call only for _parent.replaceChild(this, ...);
		@Override
		public INode23 replaceChild(INode23 oldNode, INode23 newNode){
			if(_left == oldNode){
				_left = newNode;
			} else {
				if(_middle == oldNode){
					_middle = newNode;
				} else {
					_right = newNode;
				}
			}
			return newNode;
		}
		
		@Override 
		public String toString(){
			return "["+_left+","+_min+"/"+_middle+"/"+_max+","+_right+"]";
		}
	}
}

public class TwoThreeTreeSample {
	
	public static void main(String[] args) {
		out.println("2-3 Tree Java Demo");
		TwoThreeTree tree = new TwoThreeTree();
		for(int i=1; i<10; i++){
			tree.insert(i);
			out.println(tree);
		}
		out.println("\nend");
		out.println(tree);
	}
}
